#!/usr/bin/perl -w
#
#   Copyright 2012 Andrii Golovynskyi
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

use strict;

use DBI;

use Fcntl ':mode';
use List::Util qw[min max];


# Parsing Config Module
# Use: http://search.cpan.org/~tlinden/Config-General/
# Debian: libconfig-general-perl
# RHEL: perl-Config-General
use Config::General;

use Cwd 'abs_path';
use File::Basename;

$ENV{LANG} = "C";

# Global variable 
our $prefix;

# Begin parsing config
my $configfile = new Config::General("$prefix/etc/grid.conf");
my %config = $configfile->getall;

use Date::Parse;
# RHEL: perl-TimeDate.noarch

if ((defined $config{GRID_PREFIX}) && $config{GRID_PREFIX} ne "") {
  $config{GRID_PREFIX} .= "/";
} else {
  $config{GRID_PREFIX} = "";
}

if ((defined $config{GLOBUS_PREFIX}) && $config{GLOBUS_PREFIX} ne "") {
  $config{GLOBUS_PREFIX} .= "/";
} else {
  $config{GLOBUS_PREFIX} = "";
}

if ((defined $config{LCG_PREFIX}) && $config{LCG_PREFIX} ne "") {
  $config{LCG_PREFIX} .= "/";
} else {
  $config{LCG_PREFIX} = "";
}



# --------------------------- Functions ---------------------------------------------------

sub grid_cert_install {
  # cert - certificate
  my ($cert,$key) = @_;
  mkdir "$ENV{HOME}/.globus";
  if ((defined $cert) && $cert ne "") {
    open (CERT, ">", "$ENV{HOME}/.globus/usercert.pem");
    print CERT $cert;
    close CERT;
    chmod 0400, "$ENV{HOME}/.globus/usercert.pem";
  }

  if ((defined $key) && $key ne "") {
    open (KEY, ">", "$ENV{HOME}/.globus/userkey.pem");
    print KEY $key;
    close KEY;
    chmod 0400, "$ENV{HOME}/.globus/userkey.pem";
  }
  return grid_cert_diag();
}


sub grid_vo_install {
  # Install virtual organization list
  my ($conf) = @_;
  mkdir  "$ENV{HOME}/.glite";
  open (VO, ">", "$ENV{HOME}/.glite/vomses");
  print VO $conf;
  close VO;
  return grid_vo_list();
}

sub grid_vo_list {
  # List names and contence of virtual organizations files
  my @vo_l;
  open (VO, "$ENV{HOME}/.glite/vomses");
  foreach my $l (<VO>) {
    my %vo_h;
    ($vo_h{alias},$vo_h{server},$vo_h{port},$vo_h{dn},$vo_h{name}) = split(" ", $l);
    push @vo_l, \%vo_h if defined $vo_h{alias};
  }
  close VO;
  return \@vo_l;
}


sub grid_proxy_init {
  # passwd - certificate password
  my ($passwd,$vo,$my_U,$my_H) = @_;
  my $vo_str;
  if ((defined $vo) && ($vo ne "")) {
    $vo_str = "--voms $vo";
  }

  if ((defined $my_U) && ($my_U ne "") && (defined $my_H) && ($my_H ne "")) {
      system( "echo $passwd | $config{GLOBUS_PREFIX}myproxy-logon --pshost $my_H --username $my_U --stdin_pass $vo_str >/dev/null")
        or die "Cannot init proxy!: $!\n";
  } else {
      system("echo $passwd | $config{GRID_PREFIX}voms-proxy-init -pwstdin $vo_str > /dev/null")
        or die "Cannot init proxy!: $!\n";
  }
  return grid_proxy_info();
}

sub grid_proxy_info {
  my $res;
  my %proxy;
  require "$prefix/libexec/modules/utils.pl";
  my @res = `$config{GRID_PREFIX}voms-proxy-info --all 2>&1` or return;
  my $vo_flag=0;
  foreach my $l (@res) {
    $vo_flag=1 if $l =~ m/^=== VO /;
    if (my ($var,$val) = $l =~ m/^ *([^: ]+) *: (.*)/) {
      if (not $vo_flag) {
          if ( $var eq "timeleft") {
            $proxy{$var} = str2sec($val);
          } else {
            $proxy{$var} = $val;
          }
      } else {
          if ( $var eq "timeleft") {
            $proxy{"vo_".$var} = str2sec($val);
          } else {
            $proxy{"vo_".$var} = $val;
          }
      }
    }
  }
  return \%proxy;
}

sub grid_proxy_destroy {
  my $res;
  $res = `$config{GRID_PREFIX}voms-proxy-destroy`;
  return $res;
}


# check certificate
sub grid_cert_diag {
  open (CERT, "openssl x509 -in $ENV{HOME}/.globus/usercert.pem -text -noout 2>/dev/null |") or
  open (CERT, "$ENV{HOME}/.globus/usercert.pem") or return 0;
    my $res = "";
    my $stop = 0;
    my %cert;
    foreach my $l (<CERT>) {
      $stop = 1 if ($l =~ m/--BEGIN CERTIFICATE--/);
      last if $stop;
      if (my ($p) = $l =~ m/Issuer: (.+)/) {
        $cert{'issuer'} = $p;
        next;
      }
      if (my ($p) = $l =~ m/Subject: (.+)/) {
        $cert{'person'} = $p;
        next;
      }
      if (my ($p) = $l =~ m/Not Before: (.+)/) {
        $cert{'valid_from'} = str2time($p);
        next;
      }
      if (my ($p) = $l =~ m/Not After : (.+)/) {
        $cert{'valid_to'} = str2time($p);
        next;
      }
      $cert{'key'} = 1 if -s "$ENV{HOME}/.globus/userkey.pem";
    }
    close CERT;
    return \%cert;
}

sub grid_launch {
  # Launch grif job
  # Parameters: 
  # H - cluster  @@ Not used in glite
  # J - jobname
  # d - current dir
  # iF - input files
  # oF - output files   
  # a - app name                  
  # p - partition  @@ CEID <host>:<tcp_port>/cream-<batch_system_name>-<queue_name>
  # n - number of CPUs            
  # t - time limit                        
  # r - runtime (rsl flag)
  # g - commandline args
  # mysql_host
  # mysql_login
  # mysql_pass
  # mysql_db
    my ($H,$J,$d,$iF,$oF,$a,$p,$n,$t,$r,$g,$mh,$ml,$mw,$md) = @_;
    my %res;

    #This script submit's only directly into CREAM CE queue
    #Check for correct CE
    if (not $p or $p !~ /^[\w.]+(?::\d+)?\/cream-\w+-\w+$/i) {
        $res{error} = "Bad queue name $p. There is supported only CREAM CE queue.";
        return \%res;
    }


    chdir ($d);
    my $jdl_file;
    if (defined $r && $r =~ /^rsl|jdl$/) {
        $jdl_file = $a;
    } else {
    #Generate jdl-file
    $jdl_file = "$J.jdl";
    open(JDL,">$jdl_file") or return;
    print JDL "Executable = \"$a\";\n";
    print JDL "Arguments = \"$g\";\n";
    print JDL "StdOutput = \"$J.out\";\n";
    print JDL "StdError = \"$J.err\";\n";

    #Prepare input- and output-sandbox params
    if (defined $iF && $iF ne "") {
        $iF =~ s/,/\",\"/g;
        $iF = "{ \"".$iF."\" }";
    }

    if (defined $oF && $oF ne "") {
        $oF =~ s/,/\",\"/g;
        $oF = ",\"".$oF."\"";
    }
    $oF = "{ \"$J.out\",\"$J.err\"$oF }";

    print JDL "InputSandbox = $iF;\n" if defined $iF;
    print JDL "OutputSandbox = $oF;\n";
    print JDL "OutputSandboxBaseDestURI = \"gsiftp://localhost\";\n";
    print JDL "CpuNumber = $n;\n" if defined $n;
    print JDL "CERequirements = Member(\"$r\", other.GlueHostApplicationSoftwareRunTimeEnvironment)" if defined $r;
    if ($t) {
        #FIXME: Not shure for this param of CpuTime
        my $exp_t = time() + $t+60;
        print JDL "&& other.GlueCEPolicyMaxCPUTime >= $t;\n";
        print JDL "ExpiryTime = $exp_t;\n"
    } else {
        print JDL ";\n";
        }
    } # end gen_jdl


    my $res_str;
    $res_str = `$config{GRID_PREFIX}glite-ce-job-submit --resource $p --autm-delegation \'$jdl_file\'`;
  if (my ($p) = $res_str =~ m/^(https:\/\/[\w.]+:\d+\/CREAM\d+)$/m) {
    $res{id} = $p;
    ($res{H}) = $p =~ m/\/\/(.+):/;
  } else {
    $res{error} = $res_str;
  }

  if (defined $res{id} && defined $mh) {
    # Connect to mysql
    my $dsn = "DBI:mysql:database=$md;host=$mh;port=3306";
    my $dbh = DBI->connect($dsn, $ml, $mw,{'RaiseError' => 1});

    my $q = "INSERT grid_tasks (task_id, hostname, task_name, work_dir, login, middleware) values (\"$res{id}\", \"$res{H}\", \"$J\", \"$d\", \"$ENV{USER}\", \"glite\");";
    $dbh->do($q);
    $dbh->disconnect();
  }
  return \%res;

}

#####################################################
sub grid_resources {
    my $param = shift; #Param may be: sites,services,SE,CE, default - sinfo-style
    my @attrs;
    my @result;

    if ($param eq "sites") {
    #Grid sites info
        my %sitemap;
        @attrs = qw(SiteName SiteLocation SiteDesc SiteAdmin SiteSupport SiteSecurity SiteInfo);
        parse_lcginfo("--list-site",\%sitemap,\@attrs);
        foreach (keys %sitemap) {
            push @result, \%{$sitemap{$_}};
        }
        return \@result;
    } elsif ($param eq "SE") {
    #Grid SE info
        my %semap;
        @attrs = qw(SE SEName SESite SEArch ProtType ProtVersion SAID SAName TotalOnline UsedOnline FreeOnline);
        parse_lcginfo("--list-se",\%semap,\@attrs);
        foreach (keys %semap) {
            push @result, \%{$semap{$_}};
        }
        return \@result;
    }

    #Grid Service info also need for CE info
    my %grid_srvmap;
    @attrs = qw(ServiceID ServiceType ServiceName ServiceSite ServiceVersion ServiceEndpoint
ServiceWSDL ServiceStatus ServiceStatusInfo ServiceOwner ServiceVOs);
    parse_lcginfo("--list-service",\%grid_srvmap,\@attrs);

    if ($param eq "services") {
        foreach (keys %grid_srvmap) {
            push @result, \%{$grid_srvmap{$_}};
        }
        return \@result;
    }
    
    #Grid CE info
    my %cemap;
    @attrs = qw(CE Cluster JobManager LRMS LRMSVersion CEStatus CEAppDir CEDefaultSE Priority
MaxTotalJobs MaxRunningJobs TotalJobs WaitingJobs RunningJobs AssignedJobSlots FreeJobSlots
TotalCPUs FreeCPUs MaxCPUTime EstRespTime MaxWCTime WorstRespTime CEVOs
OS OSVersion OSRelease PlatformArch CPUVendor Processor ClockSpeed
SMPSize PhysicalCPU LogicalCPU Memory VMemory TmpDir WNTmpDir CINT2000 CFP2000 InboundIP OutboundIP Tag);
    parse_lcginfo("--list-ce",\%cemap,\@attrs);
        
    # Determine user's VO
    my $proxy = grid_proxy_info();
    my $vo = "VO:".$proxy->{vo_VO} if defined $proxy->{vo_VO};
#    my (@allow_ce,@allow_se);
#    if ($vo) {
#        my $allow_ce = `$config{LCG_PREFIX}lcg-info --list-ce --sed --vo $vo`;#
#        @allow_ce = split(' ',$allow_ce);
#    }

    #CE manipulation
    foreach my $ce_key (keys %cemap) {
        if ($cemap{$ce_key}{CEStatus} eq 'Production' ) {
            $cemap{$ce_key}{UserAvail} = 'up';
        } else {
            $cemap{$ce_key}{UserAvail} = 'down';
        }
        #Set user's vo availability
        my @ce_vos = split(' ',$cemap{$ce_key}{CEVOs});
        if (not $vo or not grep(/^$vo$/,@ce_vos)) {
            $cemap{$ce_key}{UserAvail} = 'down';
        }

        #Map CE to grid site. 
        #Its's not available by default to determine CE's ownership
        $cemap{$ce_key}{CESite} = '_UNDEF_';
        my $ce_host_port = (split(/\//,$ce_key))[0];
        foreach my $srv_key (keys %grid_srvmap) {
            if ($grid_srvmap{$srv_key}{ServiceEndpoint} =~ m|^\w+://$ce_host_port/?| and
            $grid_srvmap{$srv_key}{ServiceType} =~ m/^(?:org.glite.ce.CREAM|org.edg.gatekeeper|ARC-CE)$/i) {
                $cemap{$ce_key}{CESite} = $grid_srvmap{$srv_key}{ServiceSite};
                $cemap{$ce_key}{CEService} = $grid_srvmap{$srv_key}{ServiceName};
                #Map service status to CE
                if ($grid_srvmap{$srv_key}{ServiceStatus} ne 'OK') {
                    $cemap{$ce_key}{Avail} = 'fail';
                }
                last;
            }
        }
    }
    if ($param eq "CE") {
        #Return site-info array
        foreach (keys %cemap) {
            push @result, \%{$cemap{$_}};
        }
    } else {
#sinfo-style output
#t state (Alloc,Idle)
#C count cpus
#H host
#a queue avail
#l max wall time
#D number of nodes
#P queue
#m memory on node
        foreach (keys %cemap) {
            my %ce_i = (); my %ce_a = ();
            ($ce_i{P}) = $cemap{$_}{CE} =~ /.+\/(.+)/;
            $ce_i{a} = $cemap{$_}{UserAvail};
            $ce_i{H} = $cemap{$_}{Cluster}; #Maybe use of CEsite value
            $ce_i{l} = $cemap{$_}{MaxCPUTime};
            $ce_i{D} = $cemap{$_}{PhysicalCPU}; #Sockets, not nodes
            $ce_i{m} = $cemap{$_}{Memory};
            $ce_i{C} = $cemap{$_}{FreeCPUs};
            $ce_i{t} = "idle"; #idle queue
            push @result,\%ce_i;
            
            my $alloc_cpus = $cemap{$_}{TotalCPUs} - $cemap{$_}{FreeCPUs};
            if ($alloc_cpus) {
            %ce_a = %ce_i;
                $ce_i{C} = $alloc_cpus;
                $ce_i{t} = "alloc"; #allock queue
                push @result,\%ce_a;
            }
        }
    }

return \@result;
}

### Helper function for run and parse lcg-info output
sub parse_lcginfo {
    my $info_cmd = shift;
    my $hash_ref = shift;
    my $attr_ref = shift;
    #Optional
    my $query_str = shift;
    $query_str = "--query ".$query_str if $query_str;
    $query_str ||= "";
    my $top_bdii = "--is ".$config{TOP_BDII} if $config{TOP_BDII};
    $top_bdii ||= "";
    open(LCGINFO,"$config{LCG_PREFIX}lcg-info $info_cmd --sed --attrs ".join(',', @{$attr_ref})." $query_str $top_bdii |")
        or die "Could not run lcg-info\n";
    
    foreach my $line (<LCGINFO>) {
        my @fields = split(/%/,$line);
        my $key = shift @fields;
        my %fields;
        for my $i (0 .. $#fields) {
            $fields[$i] =~ tr/&/\n/;
            $fields{$attr_ref->[$i]} = $fields[$i];
        }
        $hash_ref->{$key} = \%fields;
    }
    close LCGINFO;
}

#####################################################
#result hash of jobs
#i jobid
#j jobname
#t status
#r reason
#H cluster(host)
#u - CN username
#P - queue
#C - cpu count
#b - begin time
#e - end end time
sub grid_queue {
    # parameter detail - detailed view
    my ($detail) = @_;
    $detail = 0 if not defined $detail;

    #Obtain list of CREAM CE endpoints;
    my %cream_info;
    my @cream_epts;
    my $query="ServiceType=org.glite.ce.CREAM,ServiceStatus=OK";
    parse_lcginfo("--list-service",\%cream_info,["ServiceEndpoint"],$query);
    foreach my $key (keys %cream_info) {
        push(@cream_epts, $1) if $cream_info{$key}{ServiceEndpoint} =~ /^https:\/\/([\w.]+:\d+)\//;
    }

    my $params = "--all --level 2  --status REGISTERED:PENDING:IDLE:RUNNING:REALLY-RUNNING:HELD";
    if ($detail != 0) {
	$params .= ":DONE-OK:DONE-FAILED:CANCELLED:ABORTED:UNKNOWN";
    }
    #Fetch all jobs info from each enpoint
    my @result; my %job;
    my @res;
    foreach my $ept (@cream_epts) {
        @res=`$config{GRID_PREFIX}glite-ce-job-status -e $ept $params`;
        #parse result output
	my $proxyinfo=0; my $skip=0;
        foreach my $line (@res) {
    	    if ($line =~ /^\*{6}  JobID=\[(.+)\]/) {
    		if ($job{i}) { #Previous Job Info
    		    #prepare some mandatory values
                    $job{'t'} = 'CD'  if ($job{'Current Status'} eq 'DONE-OK');
                    $job{'t'} = 'F'   if ($job{'Current Status'} =~ /^(DONE-FAILED|ABORTED)$/);
                    $job{'t'} = 'PD'  if ($job{'Current Status'} =~ /^(REGISTERED|PENDING|IDLE)$/);
                    $job{'t'} = 'R'   if ($job{'Current Status'} =~ /^(RUNNING|REALLY-RUNNING)$/);
                    $job{'t'} = 'CA'  if ($job{'Current Status'} eq 'CANCELLED');
                    $job{'t'} = 'S'  if ($job{'Current Status'} eq 'HELD');
                    $job{'t'} = 'CG' if ($job{'Current Status'} eq 'UNKNOWN');
                    #$job{'t'} = 'TO' #Glite dostn't have this status
                    $job{r} = $job{FailureReason} || $job{Description};
                    ($job{u}) = $job{DelegProxyInfo} =~ /Holder Subject\s+:.+\/CN=(.+)/;
                    my ($queue) = $job{JDL} =~ /QueueName = "(\w+)";/;
                    my ($lrms) = $job{JDL} =~ /BatchSystem = "(\w+)";/;
                    ($job{P}) = "cream-$lrms-$queue";
                    ($job{C}) = $job{JDL} =~ /CpuNumber = (\d+);/; #May be undefined
                    $job{C} ||=1;
                    ($job{j}) = $job{JDL} =~ /Executable = "(.+?)";/; #There are no jobname field in glite
                    $job{j} =~ s/\/.*\/(.*)$/$1/;                      
    		    
        	    my %job_link = %job;
                    push @result, \%job_link;
                    $skip=0;
                }
                %job = ();
                $job{i} = $1;
                $job{H} = $ept;
            } elsif ($skip) {
        	next;
            #parse multilined proxyinfo block
            } elsif ($line =~ /^\s+DelegProxyInfo\s+=\s+\[(.+)$/) {
        	$proxyinfo=1;
        	$job{DelegProxyInfo} = $1;
	    } elsif ($proxyinfo) {
                if ($line =~ /^\s+\]$/) {
                    $proxyinfo=0;
                    
                } else {
            	    $line =~ /^\s+(.*)/;
            	    $job{DelegProxyInfo} .= "\n".$1;
                }
	    #parse all other values
            } elsif ($line =~ /^\s+([\w\s]+?)\s+=\s+\[(.*)\]$/) {
                $job{$1} = $2;
            #Get values for start and end of job
            } elsif ($line =~ /^\s+Status\s+=\s+\[REGISTERED\]\s+-\s+\[.+\]\s+\((\d+)\)$/) {
        	$job{b} = $1; #FIXME: Maybe use time of real starting job in LRMS (RUNNING) but it's unknown if failed submit
            } elsif ($line =~ /^\s+Status\s+=\s+\[(?:DONE-OK|DONE-FAILED|CANCELED|ABORTED)\]\s+-\s+\[.+\]\s+\((\d+)\)$/) {
        	$job{e} = $1;
            #skip non-usable output
            } elsif ($line =~ /Issued Commands:/) {
        	$skip=1;
            }
        }
    }
    return \@result;

}

sub grid_get {
  my($d,$jid) = @_;
  my $res = `$config{GRID_PREFIX}glite-ce-job-output -dir $d $jid`;
}

sub grid_cancel {
  my($jid) = @_;
  my $res = `$config{GRID_PREFIX}glite-ce-job-cancel --noint $jid`;
  return $res;
}

sub grid_clean {
  my($jid) = @_;
  my $res = `$config{GRID_PREFIX}glite-ce-job-purge --noint $jid`;
  return $res;
}

#############################################


sub grid_ls {
  my ($dir) = @_;
  my @list;
  my @res = `$config{LCG_PREFIX}lcg-ls -l $dir`;
  foreach my $l (@res) {
    my %f; my $d; my $date;
    if (($f{'f'},$d,$f{'s'},$date) = $l =~ m/([^ ]+) ([^ ]+) (\d+) "(.+)"/ ) {
      $f{'s'} = int $f{'s'};
      $f{'t'} = str2time($date);
      $f{'d'} = 1;
      $f{'d'} = 0 if ( $d eq "file");
      push @list, \%f if ($f{'f'} ne "");
    }
  }
  return \@list;
}


sub grid_less {
  my($f,$o,$s) = @_;
  use File::Temp qw/ :POSIX /;
  my $tf = tmpnam();
  my $res = `$config{LCG_PREFIX}lcg-cp $f $tf 2>&1`;
  my $r = less($tf,$o,$s);

  system("rm -f $tf");
  return $r;
}


sub grid_put {
  my($f) = @_;
  use File::Temp qw/ :POSIX /;

  my $tf = tmpnam();
  open (FILE, ">", $tf);
  while (<STDIN>) {
    print FILE $_;
  }
  my $res = `$config{LCG_PREFIX}lcg-cp $tf $f 2>&1`;
  system ("rm -f $tf");
  print $res;
}


sub grid_cp {
  my($sf_ref,$ddir) = @_;
  my $res = "";
  foreach my $f (@$sf_ref) {
    my ($fn) = $f =~ m/([^\/]+)$/;
    $res .= `$config{LCG_PREFIX}lcg-cp $f $ddir/$fn 2>&1`
  }
  return $res;
}


###########################

sub grid_get_str {
  my($d,$jid) = @_;
  return "$config{GRID_PREFIX}glite-ce-job-output -dir $d $jid";
}

sub grid_clean_str {
  my($jid) = @_;
  return "$config{GRID_PREFIX}glite-ce-job-purge -noint $jid";
}

1;

###########################################################################################
use Data::Dumper;

#print Dumper(grid_proxy_info());
print Dumper(grid_resources(""));
#print Dumper(grid_queue(1));
#print Dumper(grid_launch("BLALA","testjob","/home/xand/jobs_glite","test.sh,test2.sh","test.out","/bin/bash","cream.ilt.kharkov.ua:8443/cream-pbs-edu","1","30","LCG-2","-x test.sh"));

__END__
