/*#   Copyright 2012 http://scms.pro/
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
*/

/*
 * SCMS Hydra Library
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>

#include <syslog.h>
#include <fcntl.h>

#include "utils.h"
#include "privatekey.h"


#define READ 0
#define WRITE 1

#define ERROR(args...) do{ if( verbose) { fprintf( stderr, args); } else { syslog (LOG_MAKEPRI(LOG_LOCAL1, LOG_ERR), args); } }while(0);

pid_t popen2(const char *command, int *infp, int *outfp)
{
    int p_stdin[2], p_stdout[2];
    pid_t pid;

    if (pipe(p_stdin) != 0 || pipe(p_stdout) != 0)
        return -1;

    pid = fork();

    if (pid < 0)
        return pid;
    else if (pid == 0)
    { //Child
      close(p_stdin[WRITE]);
      dup2(p_stdin[READ], READ);
      close(p_stdout[READ]);
      dup2(p_stdout[WRITE], WRITE);
      close(p_stdout[READ]);
      close(p_stdin[WRITE]);

      execl("/bin/sh", "sh", "-c", command, NULL);
      perror("execl");
      exit(1);
    }

    if (infp == NULL)
      close(p_stdin[WRITE]);
    else
      *infp = p_stdin[WRITE];

    if (outfp == NULL)
      close(p_stdout[READ]);
    else
      *outfp = p_stdout[READ];

    close(p_stdin[READ]);
    close(p_stdout[WRITE]);
    return pid;
}

/*
 * popen3()
 *
 * Runs the specified command line using a shell (SECURITY WARNING: don't put
 * user-provided input here, to avoid shell expansion), opening pipes for
 * stdin, stdout, and stderr.  If NULL is specified for a given fd, that stream
 * will be discarded.
 *
 * It is recommended to set nonblock_in to zero, nonblock_outerr to nonzero.
 */
pid_t popen3(const char *command, int *fp_in, int *fp_out, int *fp_err, int nonblock_in, int nonblock_outerr)
{
#define POPEN3_READ 0
#define POPEN3_WRITE 1

	// Based on popen2() from http://snippets.dzone.com/posts/show/1134,
	// where it was listed as freeware.  Modifications by Mike Bourgeous.
	// https://gist.github.com/1022231
	
	int p_stdin[2], p_stdout[2], p_stderr[2];
	pid_t pid;

	/* Open the three pipes */
	if(pipe(p_stdin) != 0 || pipe(p_stdout) != 0 || pipe(p_stderr) != 0) {
		ERROR("popen3: unable to open pipes for I/O");
		return -1;
	}

	/* Set the pipes to nonblocking IO if rhttps://gist.github.com/1022231equested (a good idea for single-threaded apps) */
	if(nonblock_in) {
    fcntl(p_stdin[POPEN3_WRITE], F_SETFL, fcntl(p_stdin[POPEN3_WRITE], F_GETFL) | O_NONBLOCK);
	}
	if(nonblock_outerr) {
    fcntl(p_stdout[POPEN3_READ], F_SETFL, fcntl(p_stdout[POPEN3_READ], F_GETFL) | O_NONBLOCK);
    fcntl(p_stderr[POPEN3_READ], F_SETFL, fcntl(p_stderr[POPEN3_READ], F_GETFL) | O_NONBLOCK);
	}

	pid = fork();

	if (pid < 0) {
		/* An error occurred */
		return pid;
	} else if (pid == 0) {
		/* This is the child process */

		// Close the parent-side half of the pipes
		close(p_stdin[POPEN3_WRITE]);
		close(p_stderr[POPEN3_READ]);
		close(p_stdout[POPEN3_READ]);

		// Replace the real stdin/stdout/stderr with the pipes created above
		dup2(p_stdin[POPEN3_READ], fileno(stdin));
		dup2(p_stderr[POPEN3_WRITE], fileno(stderr)); // stderr is first in case err is redirected to out (i.e. with 2>&1)
		dup2(p_stdout[POPEN3_WRITE], fileno(stdout));  


		execl("/bin/sh", "sh", "-c", command, (char *)NULL);
		ERROR("popen3: execl failed");
		exit(1);
	}

	/* This is the parent process */

	if(fp_in == NULL) {
		/* fp_in was null - the caller doesn't want to write to stdin */
		close(p_stdin[POPEN3_WRITE]);
	} else {
		/* Give the caller the file number for stdin */
		*fp_in = p_stdin[POPEN3_WRITE];
	}

	if(fp_out == NULL) {
		/* fp_out was null - the caller doesn't want to read from stdout */
		close(p_stdout[POPEN3_READ]);
	} else {
		/* Give the caller the file number for stdout */
		*fp_out = p_stdout[POPEN3_READ];
	}

	if(fp_err == NULL) {
		/* fp_err was null - the caller doesn't want to read from stderr */
		close(p_stderr[POPEN3_READ]);
	} else {
		/* Give the caller the file number for stderr */
		*fp_err = p_stderr[POPEN3_READ];
	}

	return pid;
#undef POPEN3_READ
#undef POPEN3_WRITE
}

char *gnu_dirname(char *path)
{
    char *base = strrchr(path, '/');
    if (base) {
      *base = '\0';
    }
    return path;
}


char *prefix;
char *prefix_lib;

static int ctlport = 9010;
static int sock;

static char *root_of_scit, *manager_var, *mvsbin, *mvsnodes;

static int openserver( void) {

  struct sockaddr_in addr;
  int yes = 1;

  memset( &addr, 0, sizeof( addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons( ctlport);
    
  if( (sock = socket( AF_INET, SOCK_STREAM, 0)) < 0) {
	  ERROR( "socket()");
	  return 1;
  }
    
  if( setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof( int)) < 0) {
	  ERROR( "setsockopt");
	  return 1;
  }
    
  if( bind( sock, (struct sockaddr *)&addr, sizeof( addr)) < 0) {
	  ERROR( "bind()");
	  return 1;
  }

  if( listen( sock, 5) < 0) {
	  ERROR( "listen()");
	  return 1;
  }
  return 0;
}

static void sig_CHLD( int sig) {
  int stat;
  while( waitpid( -1, &stat, WNOHANG) > 0);
} 


static void help( void) {

  fprintf( stderr, "Usage:\n\
-v   verbose mode\n");
  exit( 1);
}

static void do_exec( int fd, char *cmd) {

  char str[ 4194304];
  int l;
 
  if( verbose) fprintf( stderr, "executing: %s\n", cmd);

  int infp, outfp, errfp;
  char buf[128];
  
  int type;

// Initialize SIGCHLD handling
  signal( SIGCHLD,  child_sig_handler);

//    if (popen2(cmd, &infp, &outfp) <= 0) {
    if (popen3(cmd, &infp, &outfp, &errfp, 0, 1) <= 0) {
      snprintf( str, sizeof( cmd), "{\"Hydrad ERROR\":\"Command execution error, errno %d, %s\"}", errno, strerror( errno));
		  ERROR ( "%s", str);
      return;
    }
    
    if( verbose) fprintf( stderr, "command pipe opened\n");    
// Writing data to stdin of process 
    while( (l = read_str( fd, str, sizeof( str), &type))) {
        if (type =! TYPE_IN)
          ERROR( "Get packet of wrong type %d while transmiting stdin\n");
        if( l <= 0) 
          break;
        if( verbose) fprintf( stderr, "Get from socket: '%s'[ %d]\n", str, l);
        if( !write( infp, str,l)) {
		      ERROR ( "write failed");
          break;
        }
    }
    close(infp);
    if( verbose) fprintf( stderr, "input pipe closed\n");

    int stream;
    while( stream = do_select_errout( outfp, errfp, 0)) {
      l = read(stream, str, sizeof(str));
      if( verbose) fprintf( stderr, "Get %d from read command pipe %d, Result: %s\n", l, stream, strerror( errno));
      if( l <= 0) {
        close( stream);
        if( verbose) fprintf( stderr, "pipe %d closed, l=%d\n", stream, l);
      } else {
      // Writing data to hctl
        if( stream == outfp)
          type = TYPE_OUT;
        else if( stream == errfp)
          type = TYPE_ERR;
        else
          ERROR("Get data from incorrect stream\n");
        if( verbose) fprintf( stderr, "Get from command pipe[%d]: '%s'[ %d]\n", stream, str, l);
        if ( write_str(fd,str,l, type) <= 0) {
          ERROR( "Error writing data to socket");
        }
      }
    }

    close( outfp);
    if( verbose) fprintf( stderr, "output pipe closed\n");
    close( errfp);
    if( verbose) fprintf( stderr, "error pipe closed\n");
}

void proceed( int fd) {

    char str[ 1024], *s, cmd[ 4096];
    int len;
    struct passwd *passwd;
    int type;

    if( write_str( fd, "Yes?", 4, TYPE_CTL) != 4) {
      ERROR( "Can\'t Write \"Yes\"\n");
      goto handshake;
    }
    if( read_str( fd, str, sizeof( str), &type) != 5) {
      ERROR( "'%s'[ %d]\n", str, strlen( str));
      goto handshake;
    }
    if( strcmp( str, "hydra")) {
      ERROR( "'%s'[ %d]\n", str, strlen( str));
      goto handshake;
    }

    if( write_str( fd, HANDSHAKE1, HANDSHAKE1LEN, TYPE_CTL) != HANDSHAKE1LEN)
      goto handshake;
    if( read_str( fd, str, sizeof( str), &type) != HANDSHAKE2LEN)
      goto handshake;
    if( strcmp( str, HANDSHAKE2))
      goto handshake;

    if( !write_ok( fd))
      goto handshake;

    if( (len = read_str( fd, str, sizeof( str), &type)) < 0)
      goto handshake;

    if( memcmp( str, "user ", 5))
      goto handshake;

    s = str + 5;
    if( !(passwd = getpwnam( s))) {
      ERROR( "username %s not found in user database\n", s);
      exit( 1);
    }

    if( passwd->pw_uid < 500) {
      ERROR( "user %s uid %d is not allowed to use interface\n", s, passwd->pw_uid);
      exit( 1);
    }

    if( setuid( passwd->pw_uid)) {
      ERROR( "err setuid( %s): %s", s, strerror( errno));
      exit( 1);
    }

    if( !write_ok( fd))
        goto handshake;

    clearenv();
    setenv( "USER", passwd->pw_name, 1);
    setenv( "HOME", passwd->pw_dir, 1);
    setenv( "SHELL", passwd->pw_shell, 1);
    setenv( "PATH", "/bin:/usr/bin", 1);
    setenv( "TMP", "/tmp", 1);
    setenv( "TMPDIR", "/tmp", 1);

    chdir( passwd->pw_dir);
    
    if( read_str( fd, cmd, sizeof( cmd), &type) < 0)
        goto handshake;
    if( !write_ok( fd))
        goto handshake;
    char *exec;
    asprintf(&exec,"sh -l -c '%s/%s'",prefix_lib,cmd);

    if( verbose) fprintf( stderr, "cmd: %s, exec: %s\n", cmd, exec);
    do_exec( fd, exec);

    write_str( fd, ".DONE.", 0, TYPE_CTL);

done:
    exit( 0);
handshake:
    ERROR( "handshake failed\n");
    exit( 1);
}

int main( int ac, char **av) {

  char* ptr;
  char actualpath [1024];
  ptr = realpath(av[0], actualpath);
  char *p = strdup(actualpath);
  prefix = (char*) gnu_dirname(gnu_dirname(p));
//  fprintf( stderr, "prefix: %s\n", prefix);
//  return 0;

  asprintf(&prefix_lib, "%s/libexec",prefix);

  char *s;
  struct sockaddr_in addr;
  int addr_len = sizeof( addr), cl;
  pid_t pid;

  ac--; av++;
   
  while( ac) {
    s = *av;
    if( *s != '-') break;
    s++; av++; ac--;
    switch( *s) {
    case 'v':
      verbose++;
      continue;
    default:
      help();
    }
  }

  signal( SIGCHLD, sig_CHLD);
  signal( SIGPIPE, SIG_IGN);

    
  openserver();

  if( !verbose)
    daemon( 0, 0);
    
  while( 1) {
    if( do_select( sock, 0)) {
      return 90;
    }
	
    if( (cl = accept( sock, (struct sockaddr *)&addr, &addr_len)) < 0) {
	    ERROR( "err accept(): %s\n", strerror( errno));
	    continue;
    }

    if( !(pid = fork()))
      proceed( cl);

    close( cl);

    if( pid < 0) {
      ERROR( "fork(): %s\n", strerror( errno));
    }
  }
  return 0;
}
