/*#   Copyright 2012 http://scms.pro/
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
*/

/*
 * SCMS Hydra Library
 */


#include <syslog.h>
#include <fcntl.h>
#include <signal.h>

static int verbose = 0;

// For child exit handling
static volatile sig_atomic_t got_SIGCHLD = 0;

static void child_sig_handler(int sig) {
  got_SIGCHLD = 1;
}

int is_valid_fd(int fd)
{
  return fcntl(fd, F_GETFL) != -1 || errno != EBADF;
}

static int do_select( int fd, int timeout) {

  fd_set fdset;
  struct timeval tv;
  int t;
  t = timeout ? timeout : 5;
  tv.tv_sec = t; tv.tv_usec = 0;

  while( 1) {
    FD_ZERO( &fdset);
    FD_SET( fd, &fdset);
		    
    switch( select( fd + 1, &fdset, 0, 0, &tv)) {
    case 0:
      if( !timeout) {
        tv.tv_sec = t; tv.tv_usec = 0;
        continue;
      }
      return -1;
    case -1:
      if( errno == EINTR) 
        continue;
      if( verbose) 
        fprintf( stderr, "err select(): %s\n", strerror( errno));
      return -1;
    }
    break;
  }
  return 0;
}
																		    

static int do_select_errout (int fd1, int fd2, int timeout) {
  fd_set fdset, sigset;
  struct timeval tv;
  int t;
  t = timeout ? timeout : 0;
  tv.tv_sec = t; tv.tv_usec = 100000;

  while( 1) {
    FD_ZERO( &fdset);
    FD_ZERO( &sigset);
    if( is_valid_fd(fd1)) {
      FD_SET( fd1, &fdset);
      FD_SET( fd1, &sigset);
    }
    if( is_valid_fd(fd2)) {
      FD_SET( fd2, &fdset);
      FD_SET( fd2, &sigset);
    }
    // All streams are closed
    if( !is_valid_fd(fd1) && !is_valid_fd(fd2))
      return 0;
			    
    switch( select(FD_SETSIZE, &fdset, 0, &sigset, &tv)) {
    case 0:
      if (got_SIGCHLD) {
        got_SIGCHLD = 0;
        close(fd1);
        close(fd2);
        if( verbose) fprintf( stderr, "Get SIGCHLD, closing file descriptors\n");
      }
      if( !timeout) {
        tv.tv_sec = t; tv.tv_usec = 100000;
        continue;
      }
      return -1;

    case -1:
      if( errno == EINTR) 
        continue;
      if( verbose) 
        fprintf( stderr, "err select(): %s\n", strerror( errno));
      return -1;

    default:
      if (FD_ISSET(fd1, &sigset))
        close(fd1);
      if (FD_ISSET(fd2, &sigset))
        close(fd2);
      if (FD_ISSET(fd1, &fdset))
        return fd1;
      if (FD_ISSET(fd2, &fdset))
        return fd2;
      // Else continue loop
      continue;
    }
  }
  // else some error
  return -1;
}


#define TYPE_IN  1
#define TYPE_OUT 2
#define TYPE_ERR 3
#define TYPE_CTL 4

static int write_str( int fd, char *str, int len, int type) {
  if( verbose) fprintf( stderr, "->[%d]<%s>\n", len, str);

// Send String length first
  int bytesToSend = sizeof(len);
  int bytesWritten = 0;
  int num = 0;

  for (bytesWritten = 0; bytesWritten < bytesToSend; bytesWritten += num) {
    num = write(fd, (void *)( (uintptr_t) &len + (uintptr_t)bytesWritten ), bytesToSend - bytesWritten);
    if (num <= 0 && errno != EINTR) {
      syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_ERR),"Write stream error: %s. Error=%d", strerror(errno), errno);
      return -1;
    }
  }

// Next, Send packet type
  bytesToSend = sizeof(type);
  bytesWritten = 0;
  num = 0;

  for (bytesWritten = 0; bytesWritten < bytesToSend; bytesWritten += num) {
    num = write(fd, (void *)( (uintptr_t) &type + (uintptr_t)bytesWritten ), bytesToSend - bytesWritten);
    if (num <= 0 && errno != EINTR) {
      syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_ERR),"Write stream error: %s. Error=%d", strerror(errno), errno);
      return -1;
    }
  }

// Then, sending message body
  bytesToSend = len;
  bytesWritten = 0;
  num = 0;
       
  for (bytesWritten = 0; bytesWritten < bytesToSend; bytesWritten += num) {
    num = write(fd, (void *)( (uintptr_t)str + (uintptr_t)bytesWritten ), bytesToSend - bytesWritten);
    if (num <= 0 && errno != EINTR) {
      syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_ERR),"Write stream error: %s. Error=%d", strerror(errno), errno);
      return -1;
    }
  }
    
  return len;
}


static int read_str( int fd, char *str, int size, int *type) {
  int len;
  int bytesToRead;

// Read len correctly first
  bytesToRead = sizeof( len);
  int bytesReaden = 0;
  int num = 0;

  for (bytesReaden = 0; bytesReaden < bytesToRead; bytesReaden += num) {
    num = read(fd, (void *)( (uintptr_t) &len + (uintptr_t)bytesReaden ), bytesToRead - bytesReaden);
    if (num <= 0 && errno != EINTR) {
      syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_ERR),"Read stream error: %s. Error=%d", strerror(errno), errno);
      return -1;
    }
  }

// Second, read packet type
  int t;
  bytesToRead = sizeof( t);
  bytesReaden = 0;
  num = 0;

  for (bytesReaden = 0; bytesReaden < bytesToRead; bytesReaden += num) {
    num = read(fd, (void *)( (uintptr_t) &t + (uintptr_t)bytesReaden ), bytesToRead - bytesReaden);
    if (num <= 0 && errno != EINTR) {
      syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_ERR),"Read stream error: %s. Error=%d", strerror(errno), errno);
      return -1;
    }
  }
  *type = t;

// Then read message body of length len
  bytesToRead = len;
  bytesReaden = 0;
  num = 0;

  for (bytesReaden = 0; bytesReaden < bytesToRead; bytesReaden += num) {
    num = read(fd, (void *)( (uintptr_t)str + (uintptr_t)bytesReaden ), bytesToRead - bytesReaden);
    if (num <= 0 && errno != EINTR) {
      syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_ERR),"Read stream error: %s. Error=%d", strerror(errno), errno);
      return -1;
    }
  }
  if ( bytesReaden < bytesToRead) {
    syslog(LOG_MAKEPRI(LOG_LOCAL1, LOG_ERR),"Read stream error: read %d bytes instead of %d", bytesReaden, bytesToRead);
    return -1;
  }

  str[ len] = 0;
  if( verbose) 
    fprintf( stderr, "<-[%d]<%s>\n", len, str);
  return len;
}


static int read_ok( int fd) {
  char str[ 256];
  int type;
  return (read_str( fd, str, sizeof( str), &type) == 2 && str[0] == 'O' && str[1] == 'K');
}


static int write_ok( int fd) {
  return write_str( fd, "OK", 2, TYPE_CTL) == 2;
}

